#!/usr/bin/env python
# coding: utf-8


# ESA (C) 2000-2022
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# omgrismplot_graph.py


import os
import matplotlib.pyplot as plt
import numpy as np
import glob
import re
import sys
from astropy.io import fits
import pysas.pysasplot_utils.pysasplot_utils as sasplt
import pysas.pyutils.pyutils as pyutils
from matplotlib.backends.backend_pdf import PdfPages
from astropy.wcs import WCS
from matplotlib.colors import LogNorm



def load_files(file_path):
    """
    Reads a .txt file and returns its data as a NumPy array.
    
    Args:
        file_path: the path to the text file.
        
    Output:
        data: a 1D NumPy array.
    """

    try:
        data = np.loadtxt(file_path)
    except FileNotFoundError:
        raise FileNotFoundError('Could not open {}. Quitting...'.format(file_path))
    
    return data


def get_spec_info(src_str):
    """
    Returns the formatted number of the spectrum adn the spectrum ID.

    Args:
        src_str: the unformatted info from the Fortran task.

    Output:
        tuple: the number of the spectrum and its id.
    """
    
    if src_str is None:
        return('0, 00')

    src_str = src_str.strip().rstrip()
    n_spec, spec_id = src_str.split('_')
    n_spec = int(n_spec)
    spec_id = spec_id.rstrip().strip()

    return(n_spec, spec_id)


def make_text(fits_file, spec_src):
    """
    Prepares the text to be added to the plots.
    
    Args:
        fits_file: the input FITS file.
        n_spec: the number of the spectrum.
    
    Output:
        text: the text for the upper part of the figure.
    """
    
    try:
        with fits.open(fits_file) as ff:
            obs_id = pyutils.get_key_word(ff, 'OBS_ID')
            om_filter = pyutils.get_key_word(ff, 'FILTER')
            exposure_time = pyutils.get_key_word(ff, 'EXPOSURE')
            exp_id = pyutils.get_key_word(ff, 'EXP_ID')
            date_start = pyutils.get_key_word(ff, 'DATE-OBS')
            date_end = pyutils.get_key_word(ff, 'DATE-END')
            mode = pyutils.get_key_word(ff, 'DATAMODE')
    except FileNotFoundError:
        warnings.warn('Could not open file {}.'.format(fits_file))
        obs_id, om_filter, exposure_time, exp_id, date_start, date_end, mode = ['unknown'] * 7
    
    spec_number, spec_id = get_spec_info(spec_src)

    text = 'XXM - OM - {0} - Filename: {1}\n'.format(mode, fits_file) + \
        '- OBS ID: {0} - EXP ID: {1} - Filter: {2}\n'.format(obs_id, exp_id, om_filter) + \
        'Spec No: {0} - Spec ID: {1}\n'.format(spec_number, spec_id) + \
        'Start: {0}. Stop: {1}. Exp. time: {2} secs.\n'.format(date_start, date_end, exposure_time)

    return text


def read_coords_for_region(reg_file):
    '''
    Returns the x or y coords from a text file.

    Args:
        reg_file: the file containing the x or y coordinates for regions.

    Output:
        coords: a list of coordinates.
    '''
    
    with open(reg_file) as f:
        lines = f.readlines()
        coords = []
        for line in lines:
            c_data = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", line)
            coords.append(c_data)

    return coords


def curate_text_files(text_path):
    """
    Gets the text present in the given text file.

    Args:
        text_path: the path to the text file.

    Output:
        title, lines:
    """

    with open(text_path) as tp:
        lines = tp.readlines()[0]
    lines = lines.rstrip().strip()

    lines = lines.replace('\\x', '*')
    lines = lines.replace('\\d', '')
    lines = lines.replace('\\u', '^')

    splitted_lines = lines.split('-', 1)
    title = splitted_lines[0].strip().rstrip()
    lines = splitted_lines[1]

    if '\\n' in lines:
        l1, l2 = lines.split('\\n')
        l1 = l1.replace('\\n', '')
        l2 = l2.replace('\\n', '')
        lines = l1 + '\n' + l2
    
    
    lines = lines.strip().rstrip()

    return title, lines


def draw_image_with_regions(filename, reg_file, img_file):
    '''
    Draws an image containing information from grims-OM and it's regions if available.

    Args:
        filename: the filename for the output image.

    Output:
        None
    '''
    
    try:
        with fits.open(img_file) as f:
            data_img = f[0].data
            wcs = WCS(f[0].header)
    except FileNotFoundError:
        warnings.warn('Could not open image file {}.'.format(img_file))
        return None

    fig, ax = plt.subplots(subplot_kw = {'projection' : wcs})
    fig.set_figwidth(9)
    fig.set_figheight(20)
    
    ax.axis('off')
   
    ax.imshow(data_img, norm = LogNorm(), vmin = 1, vmax = 255, aspect = 'equal', cmap = 'binary', origin = 'lower')

    try:
        with open(reg_file) as rf:
            regions = rf.readlines()
        for reg in regions:
            polygon, x1, y1, width, height, angle, colour, text = sasplt.reg_to_list(reg)
            x1 = x1 - width / 2
            y1 = y1 - height / 2
            sasplt.plot_region_box(ax, x1, y1, width, height, angle = angle, fill = False, colour = colour, text = text, transform = None)
    except FileNotFoundError:
        warnings.warn('Could not open the region file {}.'.format(reg_file))
   
    plt.savefig(filename)

    return None


def main(filename, input_fits_file, device):
    """
    Main plotting utility for omgrismplot.

    Args:
        filename: the name of the plot file.
        input_fits_file: the FITS file in which the spectra are located.
        device: the display device.
    """

    # Reads the temporary files:
    wl_files = glob.glob('wl_*')
    wl_files.sort()
    rate_files = glob.glob('rate_*')
    rate_files.sort()
    text_files = glob.glob('title_*')
    text_files.sort()

    try:
        with open('SRCNO') as src:
            src_text = src.readlines()
    except FileNotFoundError:
        warnings.warn('Could not open the SRCNO info file.')
        src_text = None
    
    n_plots = len(wl_files)

    with PdfPages(filename) as pdf:
        for i in range(0, n_plots):
            spec_src = src_text[i]
            wl_file = wl_files[i]
            n_fig, l_lim, r_lim = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", wl_file)
            l_lim = float(l_lim)
            r_lim = float(r_lim)
            wl = load_files(wl_file)
            rate_file = rate_files[i]
            rate = load_files(rate_file)
            text_file = text_files[i]


            # plotting data
            f, (ax1, ax2) = plt.subplots(2, 1)
            #gs = gspec.GridSpec(3,1)

            #plt.Figure(figsize = (14, 9))
            ax1.set_position((0.15, 0.85, 0.6, 0.08))
            ax2.set_position((0.12, 0.2, 0.8, 0.6))
            ax2.plot(wl, rate, color = 'black')
            ax2.set_xlim(left = l_lim, right = r_lim)
            
            
            fig_text = make_text(input_fits_file, spec_src)
            title, line = curate_text_files(text_file)
            
            # Labels and info
            ax1.axis('off')
            if i == 0:
                ax1.text(0, 0, fig_text, fontsize = 6)
            else:
                #ax1.text(0, 0, line)
                pass
            
            ax2.set_title(title + ' - ' + line, fontsize = 7)
            ax2.set_xlabel('Wavelength / ' + r'$\AA$')
#           plt.suptitle(title + ' - ' + line, fontsize = 11)
            ax2.set_ylabel('Counts / s / ' + r'$\AA$')
            if device.upper() =='XW':
                if os.getenv('DISPLAY'):
                    plt.show()
                else:
                    warnings.warn('warning', 'Display not available.')
            else:
                pdf.savefig()
            plt.close() 


    # Cleaning up...
    for j in wl_files:
        os.remove(j)
    for j in text_files:
        os.remove(j)
    for j in rate_files:
        os.remove(j)
    if os.path.isfile('SRCNO'):
        os.remove('SRCNO')
    if device.upper() == 'XW':
        if os.path.isfile(filename):
            os.remove(filename)


    return None


if __name__ == '__main__':
    filename = sys.argv[2]
    input_fits_file = sys.argv[1]
    device = sys.argv[3]
    device = device.replace('/', '')
    rotimg = sys.argv[4].split('=')[1]
    try:
        regplotfile = sys.argv[5].split('=')[1]
    except IndexError:
        regplotfile = ''
    try:
        specregfile = sys.argv[6].split('=')[1]
    except IndexError:
        specregfile = ''
    filename_ext = os.path.splitext(filename)[1]
    filename = filename.replace(filename_ext, '.pdf')
    main(filename, input_fits_file, device)
    
    if rotimg != '':
        regplotfile_ext = os.path.splitext(regplotfile)[1]
        regplotfile = regplotfile.replace(regplotfile_ext, '.pdf')
        draw_image_with_regions(regplotfile, specregfile, rotimg)
